# -*- coding: utf-8 -*-
"""
This module handles the data processing prior to the beam analysis.
"""

# =============================================================================
# Imports
# =============================================================================
import pandas as pd
import numpy as np
from sklearn import mixture
from scipy import stats


def raw_data(fullPath):
    """
    `raw_data` returns the power density distribution.

    Parameters
    ----------
    fullPath : str
        full path to the .csv file that contains the power density
        distribution.

    Returns
    -------
    dataframe
        power density distribution.

    """

    # Read the .csv file using tab as separator. The first row is the header,
    # therefore is it skipped
    return pd.read_csv(fullPath, header=None, sep='\t', skiprows=1)


def raw_header(fullPath):
    """
    `raw_header` returns the header of the power density distribution. The
    header contains important information regarding the beam profiling, such as
    the measurement windows, number of pixels, null point, and so on.

    Parameters
    ----------
    fullPath : str
        full path to the .csv file that contains the power density
        distribution.

    Returns
    -------
    dataframe
        header of the power density distribution.

    """

    # Read the .csv file using tab as separator. Only the first row is read
    return pd.read_csv(fullPath, header=None, sep='\t', nrows=1)


def remove_background(raw_data, raw_header):
    """
    `remove_background` returns the noise-corrected power density distribution.
    Background noise and digitizer baseline are known to negatively affect the
    soundness of the beam analysis. Therefore, a noise correction method must
    be applied before carrying on the calculations.

    ISO 13694 defines two background correction methods based on the dark-field
    map of the detector:
    1) Background map subtraction: the background map is subtracted from the
    power density distribution pixel by pixel.
    2) Average background map subtraction: the average background map is
    subtracted from the power density distribution.

    This method implements option 2), and the average background map is
    retrieved from the header.

    Parameters
    ----------
    raw_data : dataframe
        power density distribution.
    raw_header : dataframe
        header of the power density distribution.

    Returns
    -------
    dataframe
        noise corrected power density distribution.

    """

    # get_nullPoint returns the null point, that is the average background map
    return raw_data - get_nullPoint(raw_header)


def get_nullPoint(raw_header):
    """
    `get_nullPoint` returns the average background map for noise correction.

    Parameters
    ----------
    raw_header : dataframe
        header of the power density distribution.

    Returns
    -------
    float64
        null point, that is the average background map.

    """

    return raw_header.iloc[0][17]


def get_xWindow(raw_header):
    """
    `get_xWindow` returns the measurement window size on the x-axis in
    millimeter. Depending on the arrangement of the measurement setup, the
    x-axis of the detector and of the dataframe may not match.

    Parameters
    ----------
    raw_header : dataframe
        header of the power density distribution.

    Returns
    -------
    float64
        measurement window size on the x-axis in millimeter.

    """

    return raw_header.iloc[0][14]


def get_yWindow(raw_header):
    """
    `get_yWindow` returns the measurement window size on the y-axis in
    millimeter. Depending on the arrangement of the measurement setup, the
    y-axis of the detector and of the dataframe may not match.

    Parameters
    ----------
    raw_header : dataframe
        header of the power density distribution.

    Returns
    -------
    float64
        measurement window size on the y-axis in millimeter.

    """

    return raw_header.iloc[0][11]


def get_xPixel(raw_header):
    """
    `get_xPixel` returns the number of pixels on the x-axis. Depending on the
    arrangement of the measurement setup, the x-axis of the detector and of the
    dataframe may not match.

    Parameters
    ----------
    raw_header : dataframe
        header of the power density distribution.

    Returns
    -------
    int64
        the number of pixels on the x-axis.

    """

    return raw_header.iloc[0][5]


def get_yPixel(raw_header):
    """
    `get_yPixel` returns the number of pixels on the y-axis. Depending on the
    arrangement of the measurement setup, the y-axis of the detector and of the
    dataframe may not match.

    Parameters
    ----------
    raw_header : dataframe
        header of the power density distribution.

    Returns
    -------
    int64
        the number of pixels on the y-axis.

    """

    return raw_header.iloc[0][8]


def get_xResolution(raw_header):
    """
    `get_xResolution` returns the pixel resolution on the x-axis in millimeter
    per pixel. The pixel resolution is calculated by dividing the windows size
    on the x-axis by the number of pixels on the x-axis.

    Parameters
    ----------
    raw_header : dataframe
        header of the power density distribution.

    Returns
    -------
    float64
        pixel resolution about the x-axis in millimeter per pixel.

    """

    return get_xWindow(raw_header) / get_xPixel(raw_header)


def get_yResolution(raw_header):
    """
    `get_yResolution` returns the pixel resolution on the y-axis in millimeter
    per pixel. The pixel resolution is calculated by dividing the windows size
    on the y-axis by the number of pixels on the y-axis.

    Parameters
    ----------
    raw_header : dataframe
        header of the power density distribution.

    Returns
    -------
    float64
        pixel resolution about the y-axis in millimeter per pixel.

    """

    return get_yWindow(raw_header) / get_yPixel(raw_header)


def image_moments(raw_data, raw_header, p, q, x0, y0):
    """
    `image_moments` returns the nth-order of the power density distribution.
    According to Wikipedia, "In image processing, computer vision and related
    fields, an image moment is a certain particular weighted sum (moment) of
    the image pixels' intensities, or a function of such moments, usually
    chosen to have some attractive property or interpretation."

    The zeroth-order moment is the unweighted sum of the power density
    distribution. The first-order moment is the sum of each pixel weighted by
    its corresponding column index (if the moment is calculated about the
    x-axis) and by its corresponding row index (if the moment is calculated
    about the y-axis).

    M_p,q = sum_x[sum_y[(x-x0)^p * (y-y0)^q * f(x, y)]] is the image moment of
    order p and reference point x0 on the x-axis and of order q and refernce
    point y0 on the y-axis.

    Parameters
    ----------
    raw_data : dataframe
        noise-corrected power density distribution.
    raw_header : dataframe
        header of the power density distribution.
    p : int
        moment order on the x-axis.
    q : int
        moment order on the y-axis.
    x0 : int/float64
        reference point on the x-axis.
    y0 : int/float64
        reference point on the y-axis.

    Returns
    -------
    aux : float64
        moment of order p and reference point x0 on the x-axis and of order q
        and reference point y0 on the y-axis.

    """

    aux = 0

    # Convert input_data to numpy array
    raw_data_np = raw_data.to_numpy()

    # Iterate through all columns and rows
    for x in range(get_xPixel(raw_header) - 1):

        for y in range(get_yPixel(raw_header) - 1):

            # Apply formula
            aux = aux + (((x - x0)**p) * ((y - y0)**q) * raw_data_np[x][y])

    return aux


def normal_mixture(df, mix):
    """
    normal_mixture returns the normal mixture fit.

    Parameters
    ----------
    df : dataframe
        data to which the normal mixture will be fitted.
    mix : int
        number of normal mixtures. mix = [1, 2, 3]

    Raises
    ------
    Exception
        in case the number of mixtures of not 1, 2, or 3.

    Returns
    -------
    x : array of float64
        x-axis component of the normal mixture fit.
    y : array of float64
        y-axis component of the normal mixture fit.

    """

    # Convert the filtered dataframe to a 1d numpy array (excluding the NaN)
    array = df.values.flatten()
    array = array[np.logical_not(np.isnan(array))]

    # Filter the array to get only values located at the high-end of the
    # x-axis. The higher the cut-off percentage value, the more data is
    # filtered out
    cut_off = (array.max() - array.min()) * 0.5
    array = array[array >= cut_off]

    # Reshape the raw data for Normal Mixture Fit
    array = array.reshape(-1, 1)

    # Curve fitting
    gmm = (
        mixture.GaussianMixture(n_components=mix,
                                max_iter=1000,
                                covariance_type='full')
        .fit(array)
    )

    # Get the mean and std of gmm
    mean = gmm.means_
    std = np.sqrt(gmm.covariances_)
    prob = gmm.weights_

    # Create the x-axis for the curve fitting
    x = np.linspace(array.min(), array.max(), 10000)

    # Calculates the normal mixture based on the specified number of mix
    if mix == 3:
        y_3 = stats.norm.pdf(x, mean[2], std[2][0][0])*prob[2]
        y_2 = stats.norm.pdf(x, mean[1], std[1][0][0])*prob[1]
        y_1 = stats.norm.pdf(x, mean[0], std[0][0][0])*prob[0]
    elif mix == 2:
        y_3 = 0
        y_2 = stats.norm.pdf(x, mean[1], std[1][0][0])*prob[1]
        y_1 = stats.norm.pdf(x, mean[0], std[0][0][0])*prob[0]
    elif mix == 1:
        y_3 = 0
        y_2 = 0
        y_1 = stats.norm.pdf(x, mean[0], std[0][0][0])*prob[0]
    else:
        raise Exception("The number of mixtures should be 1, 2, or 3.")

    # Add the three normal mixtures
    y = y_1 + y_2 + y_3

    return x, y


def pre_top_hat(raw_data):
    """
    `pre_top_hat` returns a dataframe with the necessary data to calculate the
    top-hat factor and to plot the normalized energy curve.

    Parameters
    ----------
    raw_data : dataframe
        power density distribution.

    Returns
    -------
    df : dataframe
        necessary data to calculate the top-hat factor and to plot the
        normalized energy curve.

    """

    # Stack raw_data into a single column
    raw_data_stacked = raw_data.stack(level=-1, dropna=True)

    # Get the histogram count for each bin and sort the index
    hist = raw_data_stacked.value_counts().sort_index()

    # Finds the lower_limit, which is the intensity value (bin) with the
    # highest count
    lower_limit = hist.idxmax()

    # Creates a new data frame using <hist>
    df = pd.DataFrame(data={'Count': hist}, index=hist.index)

    # Add column <Energy> having the bin with highest count (<lower_limit>)
    # as zero
    df['Energy'] = df['Count'] * (df.index - lower_limit)

    # Because the bin with the highest count is set to zero, some energy values
    # are negative, therefore should be removed
    df = df[(df >= 0).all(axis=1)]

    # Since not all bins have a count, fill missing values with zeros
    df = (
        df.reindex(np.arange(min(df.index), max(df.index)+1), fill_value=0)
    )

    # Add reverse commulative sum
    df['Cumulative Energy'] = (
        df.loc[::-1, 'Energy'].cumsum()[::-1]
    )

    # Add normalized cumulative sum
    df['Normalized Cumulative Energy'] = (
        100*df['Cumulative Energy'].to_numpy() /
        df.max(axis=0)['Cumulative Energy']
    )

    # Add normalized intensity
    df['Normalized Intensity'] = (
        100*(df.index - min(df.index))/(max(df.index) - min(df.index))
    )

    return df



# -*- coding: utf-8 -*-
"""
This module handles the main operation of the beam analysis.
"""

# =============================================================================
# Imports
# =============================================================================
import sys

from beamprofiler import beam
from beamprofiler.utils import report, plot


def run(path, fileName, eta, epsilon, mix, **kwargs):
    """
    `run` runs the beam analysis.

    Parameters
    ----------
    path : str
        full path to the power density distribution file.
    fileName : str
        file name the power density distribution.
    eta : float
        upper clip-level. 0 < eta < 1.
    epsilon : float
        lower clip-level. o < epsilon < 1.
    mix : int
        number of normal mixtures used in the curve fitting. mix = [1, 2, 3].

    Other Parameters
    ----------------
    n_bins : int
        number of bins used in the histogram.
    zoom : float
        zoom of the inset image.
    x1 : float
        lower bound of the inset image on the x-axis.
    x2 : float
        upper bound of the inset image on the y-axis.
    y1 : float
        lower bound of the inset image on the y-axis.
    y2 : float
        upper bound of the inset image on the x-axis.
    z_lim : float
        upper intensity limite of the cross-section graph of the power density
        distribution.
    cross_x : float
        x-coordinate of the cross-section graph of the power density
        distribution
    cross_y : float
        y-coordinate of the cross-section graph of the power density
        distribution

    Returns
    -------
    None.

    """

    # Creates an instance of Beam
    myBeam = beam.Beam(path, fileName, eta, epsilon, mix)

    # Plots
    plot.run(path, fileName, myBeam, **kwargs)

    # Generates the report
    report.write(path, fileName, myBeam)


if __name__ == '__main__':

    # Get arguments from the command line
    path = sys.argv[1]
    fileName = sys.argv[2]
    eta = float(sys.argv[3])
    epsilon = float(sys.argv[4])
    mix = int(sys.argv[5])

    # Run
    run(path, fileName, eta, epsilon, mix)



# -*- coding: utf-8 -*-
"""
This module handles the generation of the auxiliary plots.
"""

# =============================================================================
# Imports
# =============================================================================
import os
import numpy as np

import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.patches as mpatches
from matplotlib.ticker import (MultipleLocator, FormatStrFormatter)

from mpl_toolkits.axes_grid1 import make_axes_locatable
from mpl_toolkits.axes_grid1.inset_locator import zoomed_inset_axes, mark_inset

from beamprofiler.utils import data_processing as dp


def general_plot():
    """
    `general_plot` returns a general-purpose, blank pyplot graph.

    Returns
    -------
    fig : Figure object of matplotlib.figure module
        blank matplotlib figure object.
    ax : AxesSubplot object of matplotlib.axes_subplots module
        blank matplotlib axes object.

    """

    golden_ratio = 1.618033988749895

    # Set global font style and size
    plt.rc('font', family='serif')
    plt.rcParams.update({'font.size': 6})

    fig = plt.figure(figsize=(5, 5 / golden_ratio), dpi=100)
    ax = fig.add_subplot(1, 1, 1)

    return fig, ax


def histogram(path, fileName, beam, **kwargs):
    """
    `histogram` plots the histogram of the power density distribution and the
    normal mixture fit of the top 40% of the range.

    Parameters
    ----------
    path : std
        path to where the graph will be saved.
    beam : Beam
        instance of type Beam.

    Other Parameters
    ----------------
    n_bins : int
        number of bins used in the histogram.
    zoom : float
        zoom of the inset image.
    x1 : float
        lower bound of the inset image on the x-axis.
    x2 : float
        upper bound of the inset image on the y-axis.
    y1 : float
        lower bound of the inset image on the y-axis.
    y2 : float
        upper bound of the inset image on the x-axis.

    Returns
    -------
    None.

    """

    # Check if any default value has been redefined in kwargs
    n_bins = kwargs.pop('bins', 256)
    zoom = kwargs.pop('zoom', 2)
    x1 = kwargs.pop('x1', 1600)
    x2 = kwargs.pop('x2', 2000)
    y1 = kwargs.pop('y1', 0)
    y2 = kwargs.pop('y2', 5000)

    # Get the figure and axes objects
    fig, ax = general_plot()

    # Add title and axis titles
    ax.set_title('Histogram Analysis', loc='center', pad=None)
    ax.set_xlabel('Power Density (ADC/px)')
    ax.set_ylabel('Number of counts')

    # Histogram
    array = beam.raw_data_null.to_numpy().flatten()
    hist, bins, bars = ax.hist(array, bins=n_bins, alpha=1, align='right',
                               histtype='stepfilled', label='Histogram')

    # Add an inset image
    inset = zoomed_inset_axes(ax, zoom=zoom, loc=1)
    inset.hist(array, bins=n_bins, alpha=1, histtype='stepfilled',
               align='right', label='Histogram')
    inset.set_xlim(x1, x2)
    inset.set_ylim(y1, y2)
    inset.tick_params(labelleft=False, labelbottom=False)
    mark_inset(ax, inset, loc1=3, loc2=3, fc="none", ec="0.5")

    # Get the highest histogram peak around the high-poewr density area (upper
    # 40% of the number of bins)
    low_bound = int(n_bins * 0.6)
    high_bound = n_bins
    max_count = max(hist[low_bound:high_bound])
    pdf_x, pdf_y = dp.normal_mixture(beam.raw_data_null, beam.mix)
    pdf_y = max_count * (pdf_y / pdf_y.max())
    inset.plot(pdf_x, pdf_y, 'k--', linewidth=0.7)

    # Now that the fit has been generated, it can be added to the main graph
    ax.plot(pdf_x, pdf_y, 'k--', linewidth=0.35)

    # Save and show
    fileName = os.path.splitext(fileName)[0]
    plt.savefig(os.path.join(path, fileName + ' - histogram.png'),
                bbox_inches='tight', dpi=300)
    plt.show()


def heat_map(path, fileName, beam, **kwargs):
    """
    `heat_map` plots the heat map of the power densidty distribution.

    Parameters
    ----------
    path : str
        path to where the graph will be saved..
    beam : Beam
        instance of type `Beam`.

    Other Parameters
    ----------------
    z_lim : float
        upper intensity limite of the cross-section graph of the power density
        distribution.
    cross_x : float
        x-coordinate of the cross-section graph of the power density
        distribution
    cross_y : float
        y-coordinate of the cross-section graph of the power density
        distribution

    Returns
    -------
    None.

    """

    # Check if any default value has been redefined in kwargs
    z_lim = kwargs.pop('z_lim', -1)
    cross_x = kwargs.pop('cross_x', beam.centerX * beam.xResolution)
    cross_y = kwargs.pop('cross_y', beam.centerY * beam.yResolution)

    # Get the figure and axes objects
    fig, main_ax = general_plot()

    # Create and configure the axes
    divider = make_axes_locatable(main_ax)

    # Right ax
    right_ax = divider.append_axes("right", 0.6, pad=0.2, sharey=main_ax)
    right_ax.yaxis.set_tick_params(labelleft=False)
    right_ax.set_xlabel('Intensity')
    if z_lim != -1:
        right_ax.set_xlim(right=z_lim)

    # Top ax
    top_ax = divider.append_axes("top", 0.6, pad=0.2, sharex=main_ax)
    top_ax.xaxis.set_tick_params(labelbottom=False)
    top_ax.set_ylabel('Intensity')
    if z_lim != -1:
        top_ax.set_ylim(top=z_lim)

    # Configure and plot main graph
    z = beam.raw_data.to_numpy()
    main_ax.imshow(z,
                   extent=(0,
                           dp.get_xWindow(beam.raw_header),
                           0,
                           dp.get_yWindow(beam.raw_header)),
                   interpolation='nearest',
                   cmap=cm.gist_rainbow_r,
                   origin='lower')
    main_ax.axvline(cross_x, color='k', linestyle="--", lw=0.8)
    main_ax.axhline(cross_y, color='k', linestyle="-", lw=0.8)
    main_ax.set_xlabel('x-axis (mm)')
    main_ax.set_ylabel('y-axis (mm)')

    # Create and configure the right sub ax
    y = np.mgrid[0:dp.get_yWindow(beam.raw_header):beam.yResolution]
    slice_y = 0
    try:
        # Data along the y-axis at the x-position defined by cross_x
        slice_y = (
            z[:, int(np.around(cross_x/beam.xResolution))]
        )
    except IndexError:
        print("Slice position is outside of the range. Please, check the "
              "slice position.")
    right_ax.plot(slice_y, y, color='k', linestyle="--", lw=0.5)

    # Create and configure the top sub ax
    x = np.mgrid[0:dp.get_xWindow(beam.raw_header):beam.xResolution]
    slice_x = 0
    try:
        # Data along the x-axis at the y-position defined by cross_y
        slice_x = (
            z[int(np.around(cross_y/beam.yResolution)), :]
        )
    except IndexError:
        print("Slice position is outside of the range. Please, check the "
              "slice position.")
    top_ax.plot(x, slice_x, color='k', linestyle="-", lw=0.5)

    # Save and show
    fileName = os.path.splitext(fileName)[0]
    plt.savefig(os.path.join(path, fileName + ' - heat map.png'),
                bbox_inches='tight', dpi=300)
    plt.show()


def norm_energy_curve(path, fileName, beam):

    # Get the figure and axes objects
    fig, ax = general_plot()

    df = dp.pre_top_hat(beam.raw_data)

    # Plot
    x = df['Normalized Intensity']
    y = df['Normalized Cumulative Energy']
    ax.plot(x, y, ls='-', color='blue', linewidth=0.5)

    # Fill area under the curve
    ax.fill_between(x, y, 0, color='blue', alpha=0.5)
    ax.fill_between(x, y, 100, color='gray', alpha=0.5)

    # x-axis
    ax.set_xlabel('Normalized Intensity (%)')
    plt.xlim(0, 100)
    ax.xaxis.set_major_locator(MultipleLocator(10))
    ax.xaxis.set_major_formatter(FormatStrFormatter('%d'))
    ax.xaxis.set_minor_locator(MultipleLocator(5))
    ax.xaxis.set_tick_params(width=0.5)

    # y-axis
    ax.set_ylabel('Normalized Cumulative Energy (%)')
    plt.ylim(0, 100)
    ax.yaxis.set_major_locator(MultipleLocator(10))
    ax.yaxis.set_major_formatter(FormatStrFormatter('%d'))
    ax.yaxis.set_minor_locator(MultipleLocator(5))
    ax.yaxis.set_tick_params(width=0.5)

    ax.spines['top'].set_linewidth(0.5)
    ax.spines['bottom'].set_linewidth(0.5)
    ax.spines['right'].set_linewidth(0.5)
    ax.spines['left'].set_linewidth(0.5)

    # Title
    plt.title('Energy curve', fontdict=None, loc='center', pad=None)

    # Legend
    blue_patch = (
        mpatches.Patch(color='blue', alpha=0.5,
                       label=str(round(beam.topHatFactor*100, 2))+'%')
    )
    gray_patch = (
        mpatches.Patch(color='gray', alpha=0.5,
                       label=str(round(100-beam.topHatFactor*100, 2))+'%')
    )
    ax.legend(handles=[blue_patch, gray_patch], loc='lower left')

    # Save and show
    fileName = os.path.splitext(fileName)[0]
    plt.savefig(os.path.join(path, fileName + ' - top-hat factor.png'),
                bbox_inches='tight', dpi=300)
    plt.show()


def run(path, fileName, beam, **kwargs):
    """
    `run` runs all the auxiliary plots.

    Parameters
    ----------
    path : str
        path where the graphs will be saved.
    beam : Beam
        instance of type `Beam`.
    **kwargs : dict
        optional keyword arguments. For more information, please see `utils.
        plot.histogram` and `utils.plot.heat_map`

    Returns
    -------
    None.

    """

    histogram(path, fileName, beam, **kwargs)
    heat_map(path, fileName, beam, **kwargs)
    norm_energy_curve(path, fileName, beam)



"""
`BeamProfiler` is a Python package for laser beam analysis and
characterization according to ISO 13694, ISO 11145, and other non-ISO
definitions commonly used in the industry.
"""


__author__ = """Wagno Alves Braganca Junior"""
__email__ = 'wagnojunior@gmail.com'
__version__ = '0.1.4'


# # =============================================================================
# # Imports
# # =============================================================================
# from src.beamprofiler import analysis
# from src.beamprofiler import beam

# __all__ = ['analysis']



# -*- coding: utf-8 -*-
"""
Main module of BeamProfiler.
"""

# =============================================================================
# Imports
# =============================================================================
import sys

from beamprofiler import analysis


if __name__ == '__main__':

    # Get arguments from the command line
    indir = sys.argv[1]
    fileName = sys.argv[2]
    eta = float(sys.argv[3])
    epsilon = float(sys.argv[4])
    mix = int(sys.argv[5])

    # Run
    analysis.run(indir, fileName, eta, epsilon, mix)
